/*
 * modules.c --- automatically generated by Apache
 * configuration script.  DO NOT HAND EDIT!!!!!
 */

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"

extern module core_module;
extern module so_module;
extern module http_module;
extern module systemd_module;

/*
 *  Modules which implicitly form the
 *  list of activated modules on startup,
 *  i.e. these are the modules which are
 *  initially linked into the Apache processing
 *  [extendable under run-time via AddModule]
 */
module *ap_prelinked_modules[] = {
  &core_module,
  &so_module,
  &http_module,
  &systemd_module,
  NULL
};

/*
 *  We need the symbols as strings for <IfModule> containers
 */

ap_module_symbol_t ap_prelinked_module_symbols[] = {
  {"core_module", &core_module},
  {"so_module", &so_module},
  {"http_module", &http_module},
  {"systemd_module", &systemd_module},
  {NULL, NULL}
};

/*
 *  Modules which initially form the
 *  list of available modules on startup,
 *  i.e. these are the modules which are
 *  initially loaded into the Apache process
 *  [extendable under run-time via LoadModule]
 */
module *ap_preloaded_modules[] = {
  &core_module,
  &so_module,
  &http_module,
  &systemd_module,
  NULL
};

